### Tensor Product Smoothing via Spectral Spline Representations: Simulation
### Nathaniel E. Helwig (helwig@umn.edu)
### Updated: 26 December 2023



#########***#########   SIMULATION DESIGN   #########***#########

### load packages
require(gamm4)    # ver 0.2-6
require(grpnet)   # ver 0.2

### load tpss function
source("smooth2d.R")

### define function
eta <- function(x, additive = TRUE){
  mx1 <- cos(2 * pi * (x[,1] - pi))
  mx2 <- 30 * (x[,2] - 0.6)^5
  mx12 <- 0
  if(!additive) mx12 <- sin(pi * (x[,1] - x[,2]))
  4 * ( mx1 + mx2 + mx12 )
}

### define simulation design factor
nobs <- c(1000, 2000)
df <- 5:10
nrep <- 100



#########***#########   SIMULATION ANALYSES   #########***#########

### dry run (for later timings)
set.seed(0)
x <- data <- data.frame(x1 = runif(2000), x2 = runif(2000))
y <- rep(1, 2000)
mod <- gamm4(y ~ t2(x1, x2, k = 5, bs = "cr"), data = data)
mod <- smooth2d(x, y, df = 5)

### loop thru design factors
for(add in c(TRUE, FALSE)){
  for(n in nobs){
    for(k in df){
      res <- data.frame(additive = add, rep = 1:nrep, 
                        gamm4.mse = NA, tpss4.mse = NA,
                        gamm4.mae = NA, tpss4.mae = NA,
                        gamm4.time = NA, tpss4.time = NA)
      for(i in 1:nrep){
        
        cat(paste0(ifelse(add, "add:  ", "int:  "), "n = ", n, ",   df = ", k, ",   rep = ", i), sep = "\n")
        
        set.seed(i)
        x <- cbind(x1 = runif(n), x2 = runif(n))
        fx <- eta(x, additive = add)
        y <- fx + rnorm(n)
        data <- cbind(y = y, as.data.frame(x))
        
        tic <- proc.time()
        mod <- gamm4(y ~ t2(x1, x2, k = k, bs = "cr"), data = data)
        toc <- proc.time() - tic
        res$gamm4.mse[i] <- mean((fx - mod$gam$fitted.values)^2)
        res$gamm4.mae[i] <- mean(abs(fx - mod$gam$fitted.values))
        res$gamm4.time[i] <- toc[3]
        
        tic <- proc.time()
        mod <- smooth2d(x, y, df = k)
        toc <- proc.time() - tic
        res$tpss4.mse[i] <- mean((fx - mod$fitted.values)^2)
        res$tpss4.mae[i] <- mean(abs(fx - mod$fitted.values))
        res$tpss4.time[i] <- toc[3]
        
      }
      save(res, file = paste0("simres_", ifelse(add, "add", "int"), "_n", n, "_df", k, ".rda"))
    }
  }
}



#########***#########   SIMULATION RESULTS   #########***#########

### load results
resall <- NULL
for(add in c(TRUE, FALSE)){
  for(n in nobs){
    for(k in df){
      load(paste0("simres_", ifelse(add, "add", "int"), "_n", n, "_df", k, ".rda"))
      resall <- rbind(resall, cbind(nobs = n, df = k, res))
    }
  }
}

### plot results (RMSE)
pdf(file = "fig3-sim-mse.pdf", width = 10, height = 6.5)
par(mfrow = c(2,2), mar = c(4.2, 4.2, 3, 2) + 0.1)
for(add in c(TRUE, FALSE)){
  for(n in nobs){
    subres <- subset(resall, additive == add & nobs == n)
    long <- data.frame(nobs = n, df = rep(subres$df, 2),
                       method = rep(c("gamm4", "tpss4"), each = nrep * length(df)),
                       rmse = sqrt(c(subres$gamm4.mse, subres$tpss4.mse)))
    boxplot(rmse ~ method * df, data = long, xlab = "Number of Knots", ylab = "RMSE", 
            main = paste0(ifelse(add, "additive:", "interaction:"), "   n = ", n), 
            at = c(1:2, 4:5, 7:8, 10:11, 13:14, 16:17), axes = FALSE, 
            col = rep(c("gray", "white"), length(df)), ylim = c(0.05, 0.42),
            cex.axis = 1.25, cex.lab = 1.5, cex.main = 1.75)
    axis(1, at = c(1.5, 4.5, 7.5, 10.5, 13.5, 16.5), labels = df, cex.axis = 1.25)
    axis(2, cex.axis = 1.25)
    box()
    grid(nx = NA, ny = NULL)
    legend("topright", legend = c("gamm4", "tpss4"), fill = c("gray", "white"), 
           cex = 1.25, bty = "n")
  }
}
dev.off()


### plot results (MAE)
pdf(file = "fig4-sim-mae.pdf", width = 10, height = 6.5)
par(mfrow = c(2,2), mar = c(4.2, 4.2, 3, 2) + 0.1)
for(add in c(TRUE, FALSE)){
  for(n in nobs){
    subres <- subset(resall, additive == add & nobs == n)
    long <- data.frame(nobs = n, df = rep(subres$df, 2),
                       method = rep(c("gamm4", "tpss4"), each = nrep * length(df)),
                       mae = c(subres$gamm4.mae, subres$tpss4.mae))
    boxplot(mae ~ method * df, data = long, xlab = "Number of Knots", ylab = "MAE", 
            main = paste0(ifelse(add, "additive:", "interaction:"), "   n = ", n),
            at = c(1:2, 4:5, 7:8, 10:11, 13:14, 16:17), axes = FALSE, 
            col = rep(c("gray", "white"), length(df)), ylim = c(0.05, 0.42),
            cex.axis = 1.25, cex.lab = 1.5, cex.main = 1.75)
    axis(1, at = c(1.5, 4.5, 7.5, 10.5, 13.5, 16.5), labels = df, cex.axis = 1.25)
    axis(2, cex.axis = 1.25)
    box()
    grid(nx = NA, ny = NULL)
    legend("topright", legend = c("gamm4", "tpss4"), fill = c("gray", "white"), 
           cex = 1.25, bty = "n")
  }
}
dev.off()


### plot results (time)
pdf(file = "fig5-sim-time.pdf", width = 10, height = 6.5)
par(mfrow = c(2,2), mar = c(4.2, 4.2, 3, 2) + 0.1)
for(add in c(TRUE, FALSE)){
  for(n in nobs){
    subres <- subset(resall, additive == add & nobs == n)
    long <- data.frame(nobs = n, df = rep(subres$df, 2),
                       method = rep(c("gamm4", "tpss4"), each = nrep * length(df)),
                       time = c(subres$gamm4.time, subres$tpss4.time))
    boxplot(time ~ method * df, data = long, xlab = "Number of Knots", ylab = "Runtime (sec)", 
            main = paste0(ifelse(add, "additive:", "interaction:"), "   n = ", n),
            at = c(1:2, 4:5, 7:8, 10:11, 13:14, 16:17), axes = FALSE, 
            col = rep(c("gray", "white"), length(df)), ylim = c(0, 20),
            cex.axis = 1.25, cex.lab = 1.5, cex.main = 1.75)
    axis(1, at = c(1.5, 4.5, 7.5, 10.5, 13.5, 16.5), labels = df, cex.axis = 1.25)
    axis(2, cex.axis = 1.25)
    box()
    grid(nx = NA, ny = NULL)
    legend("topleft", legend = c("gamm4", "tpss4"), fill = c("gray", "white"), 
           cex = 1.25, bty = "n")
  }
}
dev.off()